//**************************************************************************
//*                     This file is part of the                           *
//*                 Mpxplay/MMC - multimedia player.                       *
//*                  The source code of Mpxplay is                         *
//*        (C) copyright 1998-2019 by PDSoft (Attila Padar)                *
//*                http://mpxplay.sourceforge.net                          *
//*                  email: mpxplay@freemail.hu                            *
//**************************************************************************
//*  This program is distributed in the hope that it will be useful,       *
//*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *
//*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                  *
//*  Please contact with the author (with me) if you want to use           *
//*  or modify this source.                                                *
//**************************************************************************
//function: FFMPEG file handling (loadlist/open/close/read/seek)

//#define MPXPLAY_USE_DEBUGF 1
#define MPXPLAY_DEBUGOUT_SEEK stdout
#define MPXPLAY_DEBUGOUT_READ NULL // stdout

#include "ffmpgdec.h"

#ifdef MPXPLAY_LINK_INFILE_FF_MPEG

#include "mpxinbuf.h"
#include <control/control.h>
#include <diskdriv/diskdriv.h>

#define MPXPLAY_CONFIG_VIDEO_EXTRSTREAM_SUBDIRS_SEPARATOR ';'

extern char mpxplay_config_video_extstream_subdirs[MPXPLAY_STREAMTYPEINDEX_EXTSTREAMNUM][MPXINI_MAX_CHARDATA_LEN];

static const char *ffmpeg_supported_extensions_subtitle[] = {
"AQT",  // aqtitledec.c
"ASS",  // assdec.c
"JACO", // jacosubdec.c
"LRC",  // lrcdec.c
"MPL2", // mpl2dec.c
"PJS",  // pjsdec.c
"RT",   // realtextdec.c
"SMI",  // smidec.c
"SAMI", // smidec.c
"SCC",  // sccdec.c
"SRT",  // srtdec.c
"STL",  // stldec.c
"SUB",  // mpsubdec.c / subviewer1dec.c / subviewerdec.c
"SUP",  // supdec.c
"TED",  // tedcaptionsdec.c
"TXT",  // vplayerdec.c
"VTT",  // webvttdec.c
NULL
};

static const char *ffmpeg_supported_extensions_audiostreams[] = {
"AAC", "AC3", "AIF", "APE", "DTS", "FLAC", "MKA", "MP2", "MP3", "M4A", "MPC", "OGG", "OGA", "OPUS", "TAK", "THD", "TTA", "VQF", "WAV", "W64", "WMA", "WV", NULL
};

static int inffmpfile_search_supported_extension(int streamtype_select, char *original_filename_in, char *filename_noext_out, const char **found_extension, const char **language_extension)
{
	int streamtype_index = (streamtype_select < 0)? MPXPLAY_STREAMTYPEINDEX_AUDIO : streamtype_select;
	char *fn_extension, strtmp[MAX_PATHNAMELEN];

	if(!filename_noext_out)
	{
		filename_noext_out = &strtmp[0];
		language_extension = NULL;        // we can use this only with filename_noext_out argument
	}

	fn_extension = pds_getfilename_noext_from_fullname(filename_noext_out, original_filename_in);
	if(!fn_extension)
		return streamtype_select;

	if(found_extension)
		*found_extension = NULL;

	do{
		const char *supported_extension, **extension_list;
		unsigned int i;

		if(streamtype_index == MPXPLAY_STREAMTYPEINDEX_AUDIO)
			extension_list = &ffmpeg_supported_extensions_audiostreams[0];
		else if(streamtype_index == MPXPLAY_STREAMTYPEINDEX_SUBTITLE)
			extension_list = &ffmpeg_supported_extensions_subtitle[0];
		else
			continue;

		i = 0;
		do {
			supported_extension = extension_list[i++];
			if(pds_stricmp((char *)supported_extension, fn_extension) == 0)
			{
				if(found_extension)
					*found_extension = supported_extension;
				if(language_extension)
				{   // check language extension (videoname.lng.srt)
					char *le = pds_strrchr(filename_noext_out,'.');
					if(le)
					{
						unsigned int lang_len;
						le++;
						lang_len = pds_strlen(le);
						if((lang_len >= INFFMPG_LANGCODE_ISO6391_LEN) && (lang_len <= INFFMPG_LANGCODE_ISO6392_LEN) && pds_chkstr_letters(le)) // language extension is valid (like "HU" or"HUN" or "ENG")
							*language_extension = le;
					}
				}
				return streamtype_index;
			}
		}while(supported_extension);
	}while((streamtype_select < 0) && (++streamtype_index < MPXPLAY_STREAMTYPEINDEX_SEEKPREVIEW));

	return streamtype_select; // if the extension of an exact streamtype cannot be detected, we simply pass the streamtype without reporting error (usefull for video files)
}

static inline unsigned int inffmpfile_get_streamindex_start(int streamtype_index)
{
	return (INFFMPG_STREAM_INDEX_EXTERNALFILE_BEGIN + streamtype_index * INFFMPG_STREAM_INDEX_EXTERNALFILE_NB_MAX);
}

static int inffmpgfile_ioctx_file_read(void *opaque, uint8_t *readbuf, int readsize)
{
	struct ffmpeg_external_files_info_s *extfilehandler = (struct ffmpeg_external_files_info_s *)opaque;
	int retcode = AVERROR(ENOMEM);

	if(!extfilehandler || !extfilehandler->fbfs || !extfilehandler->fbds || !readbuf)
		return retcode;
	if(!readsize)
		return AVERROR(EINVAL);

	retcode = extfilehandler->fbfs->fread(extfilehandler->fbds, readbuf, readsize);
	if(retcode < readsize)
	{
		if(!extfilehandler->fbfs->eof(extfilehandler->fbds))
		{
			if(retcode > 0)
				extfilehandler->fbfs->fseek(extfilehandler->fbds, -retcode, SEEK_CUR); // !!! assume buffer seek
			retcode = AVERROR(EAGAIN);
		}
		else if(retcode == 0)
		{
			retcode = AVERROR_EOF;
		}
	}

	mpxplay_debugf(MPXPLAY_DEBUGOUT_READ,"inffmpgsub_ioctx_file_read st:%d rs:%d ret:%d", extfilehandler->streamtype_index, readsize, retcode);

	return retcode;
}

static int64_t inffmpgfile_ioctx_file_seek(void *opaque, int64_t offset, int whence)
{
	struct ffmpeg_external_files_info_s *extfilehandler = (struct ffmpeg_external_files_info_s *)opaque;
	int64_t retcode = AVERROR(ENOMEM);

	if(!extfilehandler || !extfilehandler->fbfs || !extfilehandler->fbds)
		return retcode;

	if(whence == AVSEEK_SIZE)
		retcode = extfilehandler->fbfs->filelength(extfilehandler->fbds);
	else
	{
		whence &= 0x03;
		retcode = extfilehandler->fbfs->fseek(extfilehandler->fbds, offset, whence);
	}
	mpxplay_debugf(MPXPLAY_DEBUGOUT_SEEK,"inffmpgfile_ioctx_file_seek st:%d of:%lld wh:%d ret:%lld fn:%s", extfilehandler->streamtype_index,
			offset, whence, retcode, pds_getfilename_from_fullname(extfilehandler->external_file_selected->external_filename));

	return retcode;
}

//----------------------------------------------------------------------------------------------------------------------------------

int in_ffmpfile_filelist_file_add(struct ffmpg_demuxer_data_s *ffmpi, char *filename, int streamtype_index)
{
	struct ffmpeg_external_files_info_s *extfiles;
	struct ffmpeg_external_filelist_elem_s **next_externalfile_in_chainlist;
	unsigned int streamindex_start, stream_index, streamindex_max;
	const char *supported_extension = NULL, *language_extension = NULL;
	char ff_filename_noext[MAX_PATHNAMELEN];

	if(!filename || !filename[0])
		return MPXPLAY_ERROR_INFILE_MEMORY;

	if(streamtype_index < 0)
	{
		streamtype_index = inffmpfile_search_supported_extension(streamtype_index, filename, &ff_filename_noext[0], &supported_extension, &language_extension);
		if(streamtype_index < 0)
			streamtype_index = MPXPLAY_STREAMTYPEINDEX_AUDIO; // if an unknown file is added to the streamlist, it's used as an audio stream (getting the first audio stream)
	}

	stream_index = streamindex_start = inffmpfile_get_streamindex_start(streamtype_index);
	streamindex_max = streamindex_start + INFFMPG_STREAM_INDEX_EXTERNALFILE_NB_MAX - 1;

	extfiles = &ffmpi->external_file_infos[streamtype_index];
	next_externalfile_in_chainlist = &(extfiles->external_files_chain);
	while(*next_externalfile_in_chainlist)
	{
		next_externalfile_in_chainlist = &((*next_externalfile_in_chainlist)->next_external_file);
		if(++stream_index >= streamindex_max)
			return MPXPLAY_ERROR_INFILE_FILEOPEN;
	}
	*next_externalfile_in_chainlist = pds_calloc(1, sizeof(struct ffmpeg_external_filelist_elem_s));
	if(!*next_externalfile_in_chainlist)
		return MPXPLAY_ERROR_INFILE_FILEOPEN;
	(*next_externalfile_in_chainlist)->stream_index = stream_index;
	pds_strcpy((*next_externalfile_in_chainlist)->format_name, (char *)supported_extension);
	pds_strcpy((*next_externalfile_in_chainlist)->language_code, (char *)language_extension);
	pds_strcpy((*next_externalfile_in_chainlist)->language_name, (char *)mpxplay_ffmplang_getlangname_by_langcode(language_extension));
	pds_strcpy((*next_externalfile_in_chainlist)->external_filename, filename);
	extfiles->streamtype_index = streamtype_index;
	extfiles->nb_external_files++;
	return MPXPLAY_ERROR_INFILE_OK;
}

// add multiply stream files to the primary video (called from video context menu)
int in_ffmpfile_filelist_files_add_lfcbds(struct ffmpg_demuxer_data_s *ffmpi, struct mpxplay_loadfile_callbackdata_s *mcbs)
{
	int i;

	if(!mcbs)
		return MPXPLAY_ERROR_INFILE_MEMORY;

	if(mcbs->filenames)
	{
		for(i = 0; i < mcbs->nb_filenames; i++)
		{
			if(mcbs->filenames[i])
			{
				in_ffmpfile_filelist_file_add(ffmpi, mcbs->filenames[i], -1);
				pds_free(mcbs->filenames[i]);
			}
		}
		pds_free(mcbs->filenames);
	}

	pds_free(mcbs);

	return MPXPLAY_ERROR_INFILE_OK;
}

void in_ffmpfile_filelist_load_list_by_ext(struct ffmpeg_external_files_info_s *extfiles, char *video_fullname, unsigned int streamtype_index)
{
	struct ffmpeg_external_filelist_elem_s **next_externalfile_in_chainlist = &extfiles->external_files_chain;
	unsigned int streamindex_start = inffmpfile_get_streamindex_start(streamtype_index);
	unsigned int stream_index = streamindex_start, streamindex_max = streamindex_start + INFFMPG_STREAM_INDEX_EXTERNALFILE_NB_MAX - 1;
	char path_video[MAX_PATHNAMELEN], path_ffb[MAX_PATHNAMELEN], scanfilename[MAX_PATHNAMELEN], fullname[MAX_PATHNAMELEN];
	char search_pathes[MPXINI_MAX_CHARDATA_LEN] = "", *curr_srch_path, *video_shortname;
	struct pds_find_t ffblk;

	// search for the same filename with any extension in mpxplay_config_video_extstream_subdirs (relative to video) (if video_extstream_subdirs empty, it searches in the current dir only)
	video_shortname = pds_getpath_from_fullname(path_video, video_fullname);
	pds_getfilename_any_noext_from_fullname(scanfilename, video_fullname);
	pds_strcat(scanfilename, ".*");

	if(streamtype_index < MPXPLAY_STREAMTYPEINDEX_EXTSTREAMNUM)
		pds_strcpy(search_pathes, &mpxplay_config_video_extstream_subdirs[streamtype_index][0]);
	curr_srch_path = &search_pathes[0];

	do{
		char *next_srch_path = pds_strchr(curr_srch_path,MPXPLAY_CONFIG_VIDEO_EXTRSTREAM_SUBDIRS_SEPARATOR); // pathes in mpxplay_config_video_extstream_subdirs are separated by ';'
		if(next_srch_path)                                                     // end of current search path from mpxplay_config_video_extstream_subdirs
			*next_srch_path++ = 0;
		pds_str_clean(curr_srch_path);
		pds_filename_assemble_fullname(path_ffb, path_video, curr_srch_path);  // assemble full path for current video_extstream_subdir
		curr_srch_path = next_srch_path;                                       // set the next searchpath for the next loop

		pds_filename_assemble_fullname(fullname, path_ffb, scanfilename);      // assemble full searchname (videofilename.*) for the scan of the current video_extstream_subdir
		if(mpxplay_diskdrive_findfirst(NULL, fullname, _A_NORMAL, &ffblk))     // if we haven't found any external media file in this path
			continue;

		extfiles->streamtype_index = streamtype_index;

		do{
			const char *supported_extension = NULL, *language_extension = NULL;
			char ff_filename_noext[MAX_PATHNAMELEN];

			if(pds_stricmp(ffblk.name, video_shortname) == 0)                   // if we've found the reference media file itself
				continue;                                                       // we don't put it on the ext list
			if(inffmpfile_search_supported_extension(streamtype_index, ffblk.name, &ff_filename_noext[0], &supported_extension, &language_extension) != streamtype_index) // is file format (extension) supported?
				continue;
			if(!supported_extension)
				continue;

			// store external file informations
			*next_externalfile_in_chainlist = pds_calloc(1, sizeof(struct ffmpeg_external_filelist_elem_s));
			if(*next_externalfile_in_chainlist)
			{
				(*next_externalfile_in_chainlist)->stream_index = stream_index++;
				pds_strcpy((*next_externalfile_in_chainlist)->format_name, (char *)supported_extension);
				pds_strcpy((*next_externalfile_in_chainlist)->language_code, (char *)language_extension);
				pds_strcpy((*next_externalfile_in_chainlist)->language_name, (char *)mpxplay_ffmplang_getlangname_by_langcode(language_extension));
				pds_filename_assemble_fullname(fullname, path_ffb, ffblk.name);
				pds_strcpy((*next_externalfile_in_chainlist)->external_filename, fullname);
				next_externalfile_in_chainlist = &(*next_externalfile_in_chainlist)->next_external_file;
				extfiles->nb_external_files++;
			}
		}while(!mpxplay_diskdrive_findnext(NULL, &ffblk) && (stream_index < streamindex_max));

		mpxplay_diskdrive_findclose(NULL, &ffblk);

	}while(curr_srch_path);
}

void in_ffmpfile_filelist_close(struct ffmpeg_external_files_info_s *extfiles)
{
	struct ffmpeg_external_filelist_elem_s *externalfile_in_chainlist = extfiles->external_files_chain;;

	in_ffmpfile_file_close(extfiles);

	while(externalfile_in_chainlist)
	{
		struct ffmpeg_external_filelist_elem_s *next = externalfile_in_chainlist->next_external_file;
		pds_free(externalfile_in_chainlist);
		externalfile_in_chainlist = next;
	}
}

int in_ffmpfile_filelist_search_by_language(struct ffmpeg_external_files_info_s *extfiles, char *lang_code)
{
	struct ffmpeg_external_filelist_elem_s *externalfile_in_chainlist = extfiles->external_files_chain;
	int stream_index = INFFMPG_STREAM_INDEX_DISABLE;

	if(lang_code && lang_code[0]) // search by lang_name
	{
		while(externalfile_in_chainlist)
		{
			if(pds_stricmp(externalfile_in_chainlist->language_code, lang_code) == 0)
			{
				stream_index = externalfile_in_chainlist->stream_index;
				break;
			}
			externalfile_in_chainlist = externalfile_in_chainlist->next_external_file;
		}
	}

	if(stream_index < 0) // then search stream without language
	{
		externalfile_in_chainlist = extfiles->external_files_chain;
		while(externalfile_in_chainlist)
		{
			if(!externalfile_in_chainlist->language_code[0])
			{
				stream_index = externalfile_in_chainlist->stream_index;
				break;
			}
			externalfile_in_chainlist = externalfile_in_chainlist->next_external_file;
		}
	}

	return stream_index;
}

static struct ffmpeg_external_filelist_elem_s *in_ffmpfile_filelist_search_by_streamindex(struct ffmpeg_external_files_info_s *extfiles, int stream_index)
{
	struct ffmpeg_external_filelist_elem_s *externalfile_in_chainlist = extfiles->external_files_chain;

	while(externalfile_in_chainlist)
	{
		if((stream_index < 0) || (externalfile_in_chainlist->stream_index == stream_index)) // (stream_index < 0) selects the first file
			break;
		externalfile_in_chainlist = externalfile_in_chainlist->next_external_file;
	}

	return externalfile_in_chainlist;
}

static unsigned int in_ffmpfile_file_utftexttype_detect(struct ffmpeg_external_files_info_s *extfilehandler)
{
	unsigned int utftype = 0;
	mpxp_uint8_t readbuf[4];

	if(extfilehandler->fbfs->fread(extfilehandler->fbds, readbuf, 3) == 3)
	{
		if((readbuf[0] == 0xff) && (readbuf[1] == 0xfe))
			utftype = MPXPLAY_TEXTCONV_TYPE_UTF16LE;
		else if((readbuf[0] == 0xfe) && (readbuf[1] == 0xff))
			utftype = MPXPLAY_TEXTCONV_TYPE_UTF16BE;
		else if((readbuf[0] == 0xef) && (readbuf[1] == 0xbb) && (readbuf[2] == 0xbf))
			utftype = MPXPLAY_TEXTCONV_TYPE_UTF8;
	}

	extfilehandler->fbfs->fseek(extfilehandler->fbds, 0, SEEK_SET);

	return utftype;
}

// the sync can be stream based (if any of the streams is live (http)) or it's connected to the audio (all av streams belong to non-live-streams (local / ftp) files)
static void inffmppfile_stream_avsync_decision(struct ffmpg_demuxer_data_s *ffmpi)
{
	struct ffmpeg_external_files_info_s *extfilehandler =  &ffmpi->external_file_infos[0];
	int stream_type, nb_used_streams = 0, nb_live_streams = 0;

	for(stream_type = 0; stream_type < ffmpi->streamtype_limit; stream_type++, extfilehandler++)
	{
		if(extfilehandler->external_file_avctx)
		{
			nb_used_streams++;
			if(funcbit_test(extfilehandler->control_flags, INFFMPG_FLAG_IS_LIVESTREAM)) // TODO: what should we do with the non-seekable (but non endless) files?
			{
				nb_live_streams++;
			}
		}
	}

	extfilehandler =  &ffmpi->external_file_infos[0];
	for(stream_type = 0; stream_type <= MPXPLAY_STREAMTYPEINDEX_SUBTITLE; stream_type++, extfilehandler++)
	{
		extfilehandler->avclocktime_us = INFFMPG_INVALID_PTS;
		extfilehandler->avtimestamp_us = INFFMPG_INVALID_PTS;
	}
	if((nb_used_streams > 1) && (nb_live_streams > 0))
		funcbit_enable(ffmpi->flags, INFFMPG_FLAG_INDEPENDENTSTREAMSYNC);
	else
		funcbit_disable(ffmpi->flags, INFFMPG_FLAG_INDEPENDENTSTREAMSYNC);
}

// returns the number of streams (curently 1)
int in_ffmpfile_file_open_by_streamindex(struct ffmpg_demuxer_data_s *ffmpi, unsigned int stream_type, int stream_index, AVStream **selected_avstream, AVCodec **codec_av_new)
{
	struct ffmpeg_external_files_info_s *extfilehandler = &ffmpi->external_file_infos[stream_type];
	struct ffmpeg_external_filelist_elem_s *extfileinfo;
	int block_size = 0, retcode = MPXPLAY_ERROR_INFILE_FILEOPEN, prereadbytes;
	AVInputFormat *demuxer = NULL;
	struct mpxpframe_s *frp; // !!! hack
	char *filename = NULL;
	AVFormatContext *fctx;

	in_ffmpfile_file_close(extfilehandler);

	extfileinfo = in_ffmpfile_filelist_search_by_streamindex(extfilehandler, stream_index);
	if(!extfileinfo)
		return retcode;
	extfilehandler->external_file_selected = extfileinfo;

	if(stream_type == MPXPLAY_STREAMTYPEINDEX_VIDEO) // means primary media file too
	{
		extfilehandler->fbfs = ffmpi->fbfs;
		extfilehandler->fbds = ffmpi->fbds;
		extfilehandler->control_flags = ffmpi->flags;
	}
	else // allocate Mpxplay's file handler structure (mpxinbuf, diskdriv) for secondary (parallel/external) files
	{
		extfilehandler->fbds = mpxplay_infile_filehandler_alloc(&extfilehandler->fbfs);
		if(!extfilehandler->fbds)
			return retcode;
		extfilehandler->control_flags = 0;
	}
	extfilehandler->streamtype_index = stream_type;

	if(!extfilehandler->fbfs->fopen(extfilehandler->fbds, extfileinfo->external_filename, ffmpi->primary_file_openmode | O_BINARY, INFFMPG_IOCTX_SIZE_MAX)){
		mpxplay_debugf(MPXPLAY_DEBUG_HEADER, "inffmpg_infile_check fopen failed");
		return MPXPLAY_ERROR_INFILE_FILEOPEN;
	}

	if(ffmpi->miis->control_cb(extfilehandler->fbds, MPXPLAY_CFGFUNCNUM_INFILE_ENTRY_ISSEEKABLE, NULL, NULL))
		funcbit_enable(extfilehandler->control_flags, INFFMPG_FLAG_IS_SEEKABLE);
	if(ffmpi->miis->control_cb(extfilehandler->fbds, MPXPLAY_CFGFUNCNUM_INFILE_ENTRY_ISSTREAM, NULL, NULL))
	{
		funcbit_enable(extfilehandler->control_flags, INFFMPG_FLAG_IS_LIVESTREAM);
		// mpxplay_mpxinbuf_buffer_check((struct mpxpframe_s *)extfilehandler->fbds, FALSE); // TODO: maybe we should read more
	}

	if(!funcbit_test(extfilehandler->control_flags, INFFMPG_FLAG_IS_LIVESTREAM) && extfilehandler->fbfs->filelength && (extfilehandler->fbfs->filelength(extfilehandler->fbds) <= 44)) // too small file
		goto err_out_open;

	// pre-load stream data
	frp = (struct mpxpframe_s *)extfilehandler->fbds;
	prereadbytes = mpxplay_diskdrive_file_config(frp->filehand_datas, MPXPLAY_DISKDRIV_CFGFUNCNUM_GET_PREREADBUFBYTES, NULL, NULL);
	if(prereadbytes > 0){
		unsigned int retry = prereadbytes / (frp->prebufferblocksize + 1);
		while((frp->prebufferbytes_forward < prereadbytes) && !mpxplay_mpxinbuf_buffer_protected_check(frp) && --retry){ pds_threads_sleep(0);}
	}

	if(stream_type == MPXPLAY_STREAMTYPEINDEX_VIDEO) // means the primary media file (can be live-stream too)
	{
		if(funcbit_test(extfilehandler->control_flags, INFFMPG_FLAG_IS_LIVESTREAM))
		{
			funcbit_disable(ffmpi->primary_file_openmode, MPXPLAY_INFILE_OPENMODE_INFO_LENGTH);
			ffmpi->miis->timemsec = 0x7eeeeeee;
		}
		filename = (ffmpi->primary_file_openmode & MPXPLAY_INFILE_OPENMODE_LOAD_PLAY)? NULL : &extfileinfo->external_filename[0];
		demuxer = (ffmpi->primary_file_openmode & MPXPLAY_INFILE_OPENMODE_AUTODETECT)? NULL : ffmpi->primary_file_demuxer;
		extfilehandler->control_flags |= (MPXPLAY_TEXTCONV_TYPE_UTF8 << INFFMPG_FLAG_TEXTTYPE_BITPOS); // FIXME: hack for internal subtitles (are they always UTF-8?)
	}
	else
	{
		if(stream_type == MPXPLAY_STREAMTYPEINDEX_SUBTITLE)  // get text coding type (utf)
		{
			extfilehandler->control_flags |= in_ffmpfile_file_utftexttype_detect(extfilehandler) << INFFMPG_FLAG_TEXTTYPE_BITPOS;
		}
	}

	// initialize FFmpeg's AVFormatContext with own ioctx
	fctx = avformat_alloc_context();
	if(!fctx)
		goto err_out_open;
	extfilehandler->external_file_avctx = fctx;

	// allocate own AVIOContext (low level file handler and buffering)
	block_size = ffmpi->miis->control_cb(extfilehandler->fbds, MPXPLAY_CFGFUNCNUM_INFILE_ENTRY_RBLOCKSIZE, NULL, NULL);
	if(block_size < INFFMPG_IOCTX_SIZE_MIN)
		block_size = INFFMPG_IOCTX_SIZE_MIN;
	else if(block_size > INFFMPG_IOCTX_SIZE_MAX)
		block_size = INFFMPG_IOCTX_SIZE_MAX;

	extfilehandler->iobuffer = av_malloc(block_size + 1024);
	if(!extfilehandler->iobuffer)
		goto err_out_open;

	extfilehandler->ioctx = avio_alloc_context(extfilehandler->iobuffer, block_size, 0, (void *)extfilehandler, inffmpgfile_ioctx_file_read, NULL, (extfilehandler->control_flags & INFFMPG_FLAG_IS_SEEKABLE)? inffmpgfile_ioctx_file_seek : NULL);
	if(!extfilehandler->ioctx)
	{
		pds_free(&extfilehandler->iobuffer);
		goto err_out_open;
	}
	fctx->pb = extfilehandler->ioctx;

	if(funcbit_test(extfilehandler->control_flags, INFFMPG_FLAG_IS_LIVESTREAM))
		fctx->format_probesize = 16384;

	if(!(ffmpi->primary_file_openmode & MPXPLAY_INFILE_OPENMODE_LOAD_SEEKTAB))
		funcbit_enable(fctx->flags, (AVFMT_FLAG_IGNIDX | AVFMT_FLAG_NOPARSE | AVFMT_FLAG_NOFILLIN));
	else if(!funcbit_test(extfilehandler->control_flags, INFFMPG_FLAG_IS_SEEKABLE))
		funcbit_enable(fctx->flags, AVFMT_FLAG_IGNIDX);
	funcbit_enable(fctx->flags, (AVFMT_FLAG_CUSTOM_IO | AVFMT_FLAG_NOBUFFER));

	if((stream_type == MPXPLAY_STREAMTYPEINDEX_AUDIO) || (stream_type == MPXPLAY_STREAMTYPEINDEX_SEEKPREVIEW))
		av_dict_set_int(&extfilehandler->external_file_avctx->metadata, "track", 0, 0);  // FIXME: disable the reading of metadata for faster open (this seems to be not working)

	retcode = avformat_open_input(&extfilehandler->external_file_avctx, filename, demuxer, NULL);
	if(retcode != 0)
		goto err_out_open;

	if(codec_av_new && selected_avstream)
	{
		enum AVMediaType required_avtype;
		AVStream *stream = NULL;
		int i;

		if(avformat_find_stream_info(fctx, NULL) < 0)
			goto err_out_open;

		required_avtype = in_ffmpstrm_streamtypeindex_to_avmediatype(stream_type);

		for(i = 0; i < fctx->nb_streams; i++) // search for the first stream
		{
			stream = fctx->streams[i];
			if(stream && stream->codecpar && (stream->codecpar->codec_type == required_avtype))
			{
				break;
			}
			stream = NULL;
		}

		if(!stream)
			goto err_out_open;

		*codec_av_new = avcodec_find_decoder(stream->codecpar->codec_id);
		if(!*codec_av_new)
			goto err_out_open;

		extfileinfo->codec_id = stream->codecpar->codec_id;
		extfileinfo->bitrate = stream->codecpar->bit_rate;
		extfileinfo->channels = stream->codecpar->channels;
		extfileinfo->samplerate = stream->codecpar->sample_rate;

		*selected_avstream = stream;

		if((stream_type == MPXPLAY_STREAMTYPEINDEX_AUDIO) || (stream_type == MPXPLAY_STREAMTYPEINDEX_SUBTITLE)) // seek (sync extra stream) to video current pos
		{
			mpxp_int64_t seek_pos_us = 0;
			if((ffmpi->avsync_timestamp >= 0) && !funcbit_test(ffmpi->flags, INFFMPG_FLAG_IS_LIVESTREAM))
				seek_pos_us = ffmpi->avsync_timestamp;
			in_ffmpfile_file_seek(extfilehandler, seek_pos_us, AVSEEK_FLAG_BACKWARD | AVSEEK_FLAG_ANY);
		}
	}

	if(fctx->duration <= 0)
		funcbit_enable(extfilehandler->control_flags, INFFMPG_FLAG_IS_STREAM); // TODO: check

	inffmppfile_stream_avsync_decision(ffmpi);

	return 1;

err_out_open:
	in_ffmpfile_file_close(extfilehandler);
	return retcode;
}

void in_ffmpfile_file_close(struct ffmpeg_external_files_info_s *extfilehandler)
{
	if(extfilehandler->external_file_avctx)
	{
		if(extfilehandler->external_file_avctx->pb)
			avio_flush(extfilehandler->external_file_avctx->pb);
		avformat_flush(extfilehandler->external_file_avctx);
		avformat_close_input(&extfilehandler->external_file_avctx);
	}
	if(extfilehandler->fbfs && extfilehandler->fbds)
		extfilehandler->fbfs->fclose(extfilehandler->fbds);
	if(extfilehandler->streamtype_index != MPXPLAY_STREAMTYPEINDEX_VIDEO)
		mpxplay_infile_filehandler_free(extfilehandler->fbds);
	extfilehandler->fbds = NULL;
	extfilehandler->fbfs = NULL;
	extfilehandler->ioctx = NULL;
	extfilehandler->iobuffer = NULL;
	extfilehandler->external_file_selected = NULL;
	extfilehandler->control_flags = 0;
	extfilehandler->avclocktime_us = INFFMPG_INVALID_PTS;
	extfilehandler->avtimestamp_us = INFFMPG_INVALID_PTS;
}

int in_ffmpfile_file_packet_read(struct ffmpeg_external_files_info_s *extfilehandler, AVPacket *pkt)
{
	int retcode = -1;

	if(!extfilehandler->external_file_avctx)
		return retcode;

	av_packet_unref(pkt);
	retcode = av_read_frame(extfilehandler->external_file_avctx, pkt);

	return retcode;
}

int in_ffmpfile_file_seek(struct ffmpeg_external_files_info_s *extfilehandler, mpxp_int64_t seek_pos_us, unsigned int seek_flags)
{
#ifdef INFFMP_USE_DEMUXER_SEEK
	struct mpxpframe_s *frp = (struct mpxpframe_s *)extfilehandler->fbds;
	unsigned int buffertype_save;
#endif
	int retcode = MPXPLAY_ERROR_INFILE_EOF;

	if(!extfilehandler->external_file_avctx)
		return retcode;
	if(funcbit_test(extfilehandler->control_flags, INFFMPG_FLAG_IS_LIVESTREAM))
		return MPXPLAY_ERROR_INFILE_OK;

#ifdef INFFMP_USE_DEMUXER_SEEK
	if(frp)
	{
		buffertype_save = frp->buffertype;
		funcbit_disable(frp->buffertype, PREBUFTYPE_INT);
	}
#endif

	retcode = av_seek_frame(extfilehandler->external_file_avctx, -1, seek_pos_us, seek_flags);
	if(retcode < 0)
	{
		funcbit_inverse(seek_flags, AVSEEK_FLAG_BACKWARD);
		funcbit_enable(seek_flags, AVSEEK_FLAG_ANY);
		retcode = av_seek_frame(extfilehandler->external_file_avctx, -1, seek_pos_us, seek_flags);
		mpxplay_debugf(MPXPLAY_DEBUGOUT_SEEK,"in_ffmpfile_file_seek FAILED st:%d pos:%lld ret:%d fn:%s", extfilehandler->streamtype_index,
				seek_pos_us, retcode, pds_getfilename_from_fullname(extfilehandler->external_file_selected->external_filename));
		if(retcode < 0)
		{
			if(extfilehandler->streamtype_index == MPXPLAY_STREAMTYPEINDEX_SUBTITLE)
			{
				retcode = av_seek_frame(extfilehandler->external_file_avctx, -1, 0, 0);
				seek_pos_us = 0;
			}
			else
			{
				if(extfilehandler->fbfs && extfilehandler->fbfs->filelength && extfilehandler->fbfs->fseek && extfilehandler->fbds)
				{
					mpxp_filesize_t seek_pos_byte;
					seek_pos_byte = (seek_flags & AVSEEK_FLAG_BYTE)? seek_pos_us : ((mpxp_int64_t)((float)extfilehandler->fbfs->filelength(extfilehandler->fbds) * (float)seek_pos_us / (float)extfilehandler->external_file_avctx->duration));
					if(extfilehandler->fbfs->fseek(extfilehandler->fbds, seek_pos_byte, SEEK_SET) == seek_pos_byte)
						retcode = MPXPLAY_ERROR_INFILE_OK;
					else
						seek_pos_us = extfilehandler->avtimestamp_us;
				}
			}
		}
	}

#ifdef INFFMP_USE_DEMUXER_SEEK
	if(frp)
	{
		funcbit_smp_copy(frp->buffertype, buffertype_save, PREBUFTYPE_INT);
	}
#endif

	retcode = (retcode < 0)? MPXPLAY_ERROR_INFILE_EOF : MPXPLAY_ERROR_INFILE_OK;

	if(retcode == MPXPLAY_ERROR_INFILE_OK)
	{
		extfilehandler->avclocktime_us = INFFMPG_INVALID_PTS;
		extfilehandler->avtimestamp_us = INFFMPG_INVALID_PTS;
	}

	mpxplay_debugf(MPXPLAY_DEBUGOUT_SEEK,"in_ffmpfile_file_seek st:%d pos:%lld ret:%d", extfilehandler->streamtype_index, seek_pos_us, retcode);

	return retcode;
}

#endif // MPXPLAY_LINK_INFILE_FF_MPEG
